/* 
**  bypass.c -- function to decide whether virus scanning can be bypassed
**
**  (c) 2003-2009 Dr. Andreas Mueller, Beratung und Entwicklung
**
**  This module is distributed under the GNU General Public License (GPL).
**  See the file COPYING for details.
**
**  $Id: bypass.c,v 1.1 2009/04/10 23:32:37 afm Exp $
*/ 
#include <mod_clamav.h>

/* function to decide whether it is safe to bypass virus checking */
int	mod_clamav_safe_to_bypass(mod_clamav_config_rec *rec,
	ap_filter_t *f) {
    int			rc = 1, i, l;
    char		*ct;
    const char		*action;
    mod_clamav_safeuri	*p;

    /* requests with only headers can be bypassed			*/
    if (f->r->header_only) {
	mod_clamav_set_bypassed_note(rec, f, "header only request");
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r, "[%d] safe to skip HEAD",
		(int)getpid());
	goto safe;
    }

    /* get the module record						*/
    if (NULL == rec) {
	mod_clamav_set_bypassed_note(rec, f, "module not configured");
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
	    "[%d] no config, assuming safe", (int)getpid());
	goto safe;
    }

    /* only GET and POST return something to the caller, so all others	*/
    /* can safely bypass the virus checker				*/
    if (strcasecmp(f->r->method, "GET") && strcasecmp(f->r->method, "POST")) {
	mod_clamav_set_bypassed_note(rec, f, apr_psprintf(f->r->pool, "%s request",
	    f->r->method));
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, f->r,
	    "[%d] bypassing secure request type: %s", (int)getpid(),
	    f->r->method);
	goto safe;
    }

    /* check safe uri */
    p = (mod_clamav_safeuri *)(rec->safeuris->elts);
    for (i = 0; i < rec->safeuris->nelts; i++) {
	switch (p[i].matchtype) {
	    case MATCH_SAFE_URI:
		mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		    "[%d] checking uri '%s' against pattern '%s'",
		    (int)getpid(), f->r->uri, p[i].pattern);
		if (ap_regexec(p[i].regex, f->r->uri, 0, NULL, 0) == 0) {
		    mod_clamav_set_bypassed_note(rec, f,
			apr_psprintf(f->r->pool, "safe uri"));
		    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
			"[%d] safe uri '%s' matched pattern '%s'",
			(int)getpid(), f->r->uri, p[i].pattern);
		    goto safe;
		}
		break;
	    case MATCH_SAFE_HOST:
		mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		    "[%d] checking host '%s' against pattern '%s'",
		    (int)getpid(), f->r->hostname, p[i].pattern);
		if (ap_regexec(p[i].regex, f->r->hostname, 0, NULL, 0) == 0) {
		    mod_clamav_set_bypassed_note(rec, f,
			apr_psprintf(f->r->pool, "safe host"));
		    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
			"[%d] safe host '%s' matched pattern '%s'",
			(int)getpid(), f->r->uri, p[i].pattern);
		    goto safe;
		}
		break;
	    default:
		mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		    "[%d] ignoring unrecognized safe match type... ",
		    (int)getpid());
	}
    }

    /* if we don't know the type, we cannot bypass			*/
    if (NULL == f->r->content_type)
	goto unsafe;

    /* if mod_dnsbl told us something, we may have no other choice	*/
    action = apr_table_get(f->r->notes, "dnsbl:scan");
    if (NULL != action) {
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] checking dnsbl:action: %s", (int)getpid(), action);
	if (NULL != strstr(action, "no")) {
	    mod_clamav_set_bypassed_note(rec, f, "told to bypass by dnsbl");
	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
		"[%d] skipping scan by dnsbl", (int)getpid());
	    goto safe;
	}
	if (NULL != strstr(action, "yes")) {
	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
		"[%d] forcing scan by dnsbl", (int)getpid());
	    goto unsafe;
	}
    }

    /* find the type							*/
    ct = (char *)apr_pstrdup(f->r->pool, f->r->content_type);
    l = strlen(ct);
    for (i = 0; i < l; i++) ct[i] = tolower(ct[i]);

    /* if the type is among the safe types, we are safe			*/
    if (apr_table_get(rec->safetypes, ct)) {
	mod_clamav_set_bypassed_note(rec, f,
	    apr_psprintf(f->r->pool, "safe types '%s'", ct));
	goto safe;
    }
unsafe:
    rc = 0;
safe:
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] bypass for %s is %s", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri,
	(rc) ? "safe" : "not safe");
    return rc;
}
