/* 
**  config.c -- configuration handling routings for mod_clamav
**
**  (c) 2003-2009 Prof. Dr. Andreas Mueller, Beratung und Entwicklung
**
**  This module is distributed under the GNU General Public License (GPL).
**  See the file COPYING for details.
**
**  $Id: config.c,v 1.2 2009/04/10 23:37:28 afm Exp $
*/ 
#include <mod_clamav.h>

/* get the module configuration record */
mod_clamav_config_rec	*mod_clamav_get_module_config(request_rec *r) {
    mod_clamav_config_rec	*result;

    /* make sure we were not handed an empty request, since we don't	*/
    /* know the server, we cannot log anywhere using Apache's API	*/
    if (NULL == r) {
	fprintf(stderr, "[%d] NULL request handed to clamav_get_module_config",
	    (int)getpid());
	return NULL;
    }

    /* try to recover the directory configuration record from the	*/
    /* the apache runtime						*/
    result = (mod_clamav_config_rec *)ap_get_module_config(
	r->per_dir_config, &clamav_module);
    if (NULL != result) {
	return result;
    }

    /* if we get to this point, then there is no module config		*/
    /* record around. This really should not happen, but it did		*/
    /* an a SuSE Linux 9.0 System with the prefork MT			*/
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "[%d] no config record",
	(int)getpid());
    return NULL;
}

/* configuration methods */
void	*mod_clamav_create_dir_config(apr_pool_t *p, char *d) {
    mod_clamav_config_rec	*cfg = (mod_clamav_config_rec *)apr_pcalloc(p,
	sizeof(mod_clamav_config_rec));
    cfg->mode = MOD_CLAMAV_LOCAL;
    cfg->dbdir = NULL;
    cfg->tmpdir = NULL;
    cfg->port = 0;
    cfg->common = apr_pcalloc(p, sizeof(mod_clamav_common));
    cfg->common->local = NULL;
    cfg->trickle_interval = 60;
    cfg->trickle_size = 1;
    cfg->reload_interval = 0;
    cfg->acceptdaemonproblem = 0;
    cfg->extendedlogging = 0;
    cfg->perms = 0640;
    cfg->sizelimit = 0;
    cfg->maxfiles = 100;
    cfg->maxfilesize = 10 * 1048576;
    cfg->maxreclevel = 8;
    cfg->safetypes = apr_table_make(p, 0);
    cfg->safepatterns = apr_array_make(p, 64, sizeof(mod_clamav_safepattern));
    cfg->safeuris = apr_array_make(p, 64, sizeof(mod_clamav_safeuri));
    cfg->message = NULL;
    cfg->mutex = NULL;
    cfg->mutexname = NULL;
    cfg->shm = NULL;
    cfg->shmname = NULL;
    cfg->pool = p;

    return cfg;
} /* clamav_create_dir_config */

void	*mod_clamav_merge_dir_config(apr_pool_t *p, void *pp, void *cp) {
    mod_clamav_config_rec	*parent = (mod_clamav_config_rec *)pp;
    mod_clamav_config_rec	*child = (mod_clamav_config_rec *)cp;

    /* allways share the common part					*/
    /* XXX what is the correct way to free memory that was apr_palloced?
    if (child->common != parent->common) {
	apr_free(child->common);
    }
    */
    child->common = parent->common;
    child->mode = parent->mode;
    child->extendedlogging = parent->extendedlogging;
    child->perms = parent->perms;
    if (parent->dbdir)
	child->dbdir = (char *)apr_pstrdup(p, parent->dbdir);
    if (parent->socket)
	child->socket = (char *)apr_pstrdup(p, parent->socket);
    if (parent->reload_interval > 0)
	child->reload_interval = parent->reload_interval;
    child->trickle_size = parent->trickle_size;
    child->trickle_interval = parent->trickle_interval;
    child->sizelimit = parent->sizelimit;

    /* merge rest (except shared memory and mutex, which are not 	*/
    /* shared)								*/
    if (child->port != parent->port) child->port = parent->port;
    if (child->shmname == NULL)
	child->shmname = parent->shmname;
    if (child->mutexname == NULL)
	child->mutexname = parent->mutexname;
    if (child->maxfiles == 0)
	child->maxfiles = parent->maxfiles;
    if (child->maxfilesize == 0)
	child->maxfilesize = parent->maxfilesize;
    if (child->maxreclevel == 0)
	child->maxreclevel = parent->maxreclevel;
    if ((!child->tmpdir) && (parent->tmpdir))
	child->tmpdir = (char *)apr_pstrdup(p, parent->tmpdir);
    if (apr_is_empty_table(child->safetypes))
	apr_table_overlay(p, parent->safetypes, child->safetypes);
    if (apr_is_empty_array(child->safepatterns))
	child->safepatterns = apr_array_copy_hdr(p, parent->safepatterns);
    if ((!child->message) && (parent->message))
	child->message = (char *)apr_pstrdup(p, parent->message);

    /* note: acceptdaemonproblem is not merged */

    /* return child configuration					*/
    return child;
} /* clamav_merge_dir_config */

/* add safe types to the list of safe types */
static const char	*mod_clamav_add_safetype(cmd_parms *parms,
	void *mconfig, const char *arg) {
    char	*argi;
    int		i, l;

    /* get the module configuration record				*/
    mod_clamav_config_rec	*rec = (mod_clamav_config_rec *)mconfig;

    /* convert the type to lower case					*/
    argi = apr_pstrdup(parms->server->process->pconf, arg);
    l = strlen(argi);
    for (i = 0; i < l; i++) argi[i] = tolower(argi[i]);
    apr_table_set(rec->safetypes, argi, "safe");
    return NULL;
} /* clamav_add_safe_type */

/* mode configuration parsing function */
static const char	*mod_clamav_set_mode_cmd(cmd_parms *parms,
	void *mconfig, const char *arg) {
    mod_clamav_config_rec	*cfg = (mod_clamav_config_rec *)mconfig;

    if (0 == strcasecmp("daemon", arg)) {
	cfg->mode = MOD_CLAMAV_DAEMON;
	return NULL;
    }
    if (0 == strcasecmp("local", arg)) {
	cfg->mode = MOD_CLAMAV_LOCAL;
	return NULL;
    }
    return "unknown Clamav Mode value";
} /* clamav_set_mode_cmd */

/* permissions configuration parsing function */
static const char	*mod_clamav_set_perms_cmd(cmd_parms *parms,
	void *mconfig, const char *arg) {
    mod_clamav_config_rec	*cfg = (mod_clamav_config_rec *)mconfig;
    
    if (1 != sscanf(arg, "%o", &cfg->perms)) {
	return "unparsable Clamav temporary file permissions value";
    }
    if (cfg->perms > 0777) {
	return "Clamav temporary file permissions value out of range";
    }
    return NULL;
} /* clamav_set_perms_cmd */

/* add a pattern for which it is safe to bypass the virus scanner */
static const char	*mod_clamav_add_safepattern(cmd_parms *parms,
	void *mconfig, const char *arg1, const char *arg2, const char *arg3) {
    mod_clamav_safepattern	*p;
    int		mode;
    int		i;
    unsigned char 	pattern[16];
    unsigned char 	mask[16];

    /* get the module configuration record                             */
    mod_clamav_config_rec  *rec = (mod_clamav_config_rec *)mconfig;

    if (strlen(arg1) >= sizeof(p->tag))
	return "Clamav Safepattern tag is too long";

    if (!arg2) {
	/* recognize the text tag */
	mode = -1;
    } else {
	for (mode = 0; mode < sizeof(pattern); mode++) {
	    if (!*arg2) {
		break;
	    } else if (*arg2 == '\\') {
		if (!arg2[1]) break;
		arg2++;

		if (*arg2 == 'x' || *arg2 == 'X') {
		    if (!arg2[1] || !arg2[2]) break;
		    arg2++;

		    if (*arg2 >= '0' && *arg2 <= '9')
			pattern[mode] = *arg2 - '0';
		    else if (*arg2 >= 'A' && *arg2 <= 'F')
			pattern[mode] = *arg2 - 'A' + 10;
		    else if (*arg2 >= 'a' && *arg2 <= 'f')
			pattern[mode] = *arg2 - 'a' + 10;
		    else break;

		    pattern[mode] <<= 4;
		    arg2++;

		    if (*arg2 >= '0' && *arg2 <= '9')
			pattern[mode] |= *arg2 - '0';
		    else if (*arg2 >= 'A' && *arg2 <= 'F')
			pattern[mode] |= *arg2 - 'A' + 10;
		    else if (*arg2 >= 'a' && *arg2 <= 'f')
			pattern[mode] |= *arg2 - 'a' + 10;
		    else break;

		    arg2++;
		} else {
		    pattern[mode] = *arg2++;
		}
	    } else {
		pattern[mode] = *arg2++;
	    }
	    mask[mode] = 0xff;
	}

	if (*arg2) return "Clamav Safepattern pattern is invalid or too long";

	if (arg3) {
	    for (i = 0; i < mode; i++) {
		if (!*arg3) {
		    break;
		} else if (*arg3 == '\\') {
		    if (!arg3[1]) break;
		    arg3++;
    
		    if (*arg3 == 'x' || *arg3 == 'X') {
			if (!arg3[1] || !arg3[2]) break;
			arg3++;
    
			if (*arg3 >= '0' && *arg3 <= '9')
			    mask[i] = *arg3 - '0';
			else if (*arg3 >= 'A' && *arg3 <= 'F')
			    mask[i] = *arg3 - 'A' + 10;
			else if (*arg3 >= 'a' && *arg3 <= 'f')
			    mask[i] = *arg3 - 'a' + 10;
			else break;
    
			mask[i] <<= 4;
			arg3++;
    
			if (*arg3 >= '0' && *arg3 <= '9')
			    mask[i] |= *arg3 - '0';
			else if (*arg3 >= 'A' && *arg3 <= 'F')
			    mask[i] |= *arg3 - 'A' + 10;
			else if (*arg3 >= 'a' && *arg3 <= 'f')
			    mask[i] |= *arg3 - 'a' + 10;
			else break;

			arg3++;
		    } else {
			mask[i] = *arg3++;
		    }
		} else {
		    mask[i] = *arg3++;
		}
	    }

	    if (*arg3) return "Clamav Safepattern mask is invalid or too long";
	}
    }

    if (mode) {
	p = apr_array_push(rec->safepatterns);
	p->mode = (char)(mode);
	strcpy(p->tag, arg1);
	memcpy(p->pattern, pattern, sizeof(pattern));
	memcpy(p->mask, mask, sizeof(mask));
    } else {
	return "Clamav Safepattern pattern is empty";
    }

    return NULL;
} /* clamav_add_safe_pattern */

/* add a uri/host pattern for which it is safe to bypass the virus scanner */
static const char	*mod_clamav_add_safeuri(cmd_parms *parms, void *mconfig,
    const char *type, const char *safeuri) {
    mod_clamav_safeuri	*p;
    ap_regex_t		*preg;

    /* get the module configuration record */
    mod_clamav_config_rec  *rec = (mod_clamav_config_rec *)mconfig;

    if (strlen(safeuri) >= sizeof(p->pattern))
	return "Clamav Safeuri pattern is too long";

    p = apr_array_push(rec->safeuris);

    if (strcasecmp(type, "host") == 0) {
	p->matchtype = MATCH_SAFE_HOST;
    } else if (strcasecmp(type, "uri") == 0) {
	p->matchtype = MATCH_SAFE_URI;
    } else {
	return "Clamav unrecognized match type";
    }

    preg = ap_pregcomp(rec->pool, safeuri, REG_EXTENDED);
    if (preg == NULL) {
	return "Clamav error compiling Safeuri regex pattern";
    }

    strcpy(p->pattern, safeuri);
    p->regex = preg;

    return NULL;
} /* clamav_add_safeuri */

/* configuration directives */
const command_rec	mod_clamav_cmds[] = {
    AP_INIT_TAKE1(
	"ClamavMode", mod_clamav_set_mode_cmd,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, mode),
	RSRC_CONF,
	"virus scanning mode: local or daemon"
    ),
    AP_INIT_TAKE1(
	"ClamavSocket", ap_set_string_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, socket),
	RSRC_CONF,
	"Socket file name if clamd uses Unix domain socket"
    ),
    AP_INIT_TAKE1(
	"ClamavDbdir", ap_set_string_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, dbdir),
	RSRC_CONF,
	"Directory containing clamav databases"
    ),
    AP_INIT_TAKE1(
	"ClamavTmpdir", ap_set_string_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, tmpdir),
	RSRC_CONF,
	"Directory for temporary files during virus checking"
    ),
    AP_INIT_TAKE1(
	"ClamavReloadInterval", ap_set_int_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, reload_interval),
	RSRC_CONF,
	"Database reload interval"
    ),
    AP_INIT_TAKE1(
	"ClamavTrickleInterval", ap_set_int_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, trickle_interval),
	RSRC_CONF,
	"Trickle interval"
    ),
    AP_INIT_TAKE1(
	"ClamavTrickleSize", ap_set_int_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, trickle_size),
	RSRC_CONF,
	"Trickle size"
    ),
    AP_INIT_TAKE1(
	"ClamavMaxfiles", ap_set_int_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, maxfiles),
	RSRC_CONF,
	"Maximum number of files in an archive"
    ),
    AP_INIT_TAKE1(
	"ClamavMaxfilesize", ap_set_int_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, maxfilesize),
	RSRC_CONF,
	"Maximum archive size"
    ),
    AP_INIT_TAKE1(
	"ClamavRecursion", ap_set_int_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, maxreclevel),
	RSRC_CONF,
	"Maximum recursion depth when analyzing archives"
    ),
    AP_INIT_TAKE1(
	"ClamavPort", ap_set_int_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, port),
	RSRC_CONF,
	"Port number where clamd can be reached"
    ),
    AP_INIT_TAKE1(
	"ClamavShm", ap_set_string_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, shmname),
	RSRC_CONF,
	"shared memory file name"
    ),
    AP_INIT_TAKE1(
	"ClamavMutex", ap_set_string_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, mutexname),
	RSRC_CONF,
	"mutex lock file name"
    ),
    AP_INIT_ITERATE(
	"ClamavSafetypes", mod_clamav_add_safetype,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, safetypes),
	RSRC_CONF,
	"a list of MIME types for which it is safe to bypass the virus checker"
    ),
    AP_INIT_ITERATE(
	"ClamavSizelimit", ap_set_int_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, sizelimit),
	RSRC_CONF,
	"files larger than the sizelimit will only be checked up to that point"
    ),
    AP_INIT_FLAG(
	"ClamavAcceptDaemonproblem", ap_set_flag_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, acceptdaemonproblem),
	RSRC_CONF,
	"wether or not to live with a misbehaving daemon"
    ),
    AP_INIT_FLAG(
	"ClamavExtendedLogging", ap_set_flag_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, extendedlogging),
	RSRC_CONF,
	"whether to use extended logging (yes|no)"
    ),
    AP_INIT_TAKE1(
	"ClamavPermissions", mod_clamav_set_perms_cmd,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, perms),
	RSRC_CONF,
	"Access permissions for temporary files (octal permission value)"
    ),
    AP_INIT_TAKE1(
	"ClamavMessage", ap_set_string_slot,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, message),
	RSRC_CONF,
	"Custom virus notification message"
    ),
    AP_INIT_TAKE123(
	"ClamavSafepattern", mod_clamav_add_safepattern,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, safepatterns),
	RSRC_CONF,
	"A file pattern for which it is safe to bypass the virus checker"
    ),
    AP_INIT_TAKE2(
	"ClamavSafeURI", mod_clamav_add_safeuri,
	(void *)APR_OFFSETOF(mod_clamav_config_rec, safeuris),
	RSRC_CONF,
	"A match type (`host' or `uri') and a pattern for which it is safe to bypass the virus checker"
    ),

    { NULL }
};

