/* 
**  filter.c -- apache output filter for mod_clamav
**
**  (c) 2003-2009 Dr. Andreas Mueller, Beratung und Entwicklung
**
**  This module is distributed under the GNU General Public License (GPL).
**  See the file COPYING for details.
**
**  $Id: filter.c,v 1.1 2009/04/10 23:32:37 afm Exp $
*/ 
#include <mod_clamav.h>

/* initialize the virus filter */
static int	mod_clamav_initialize(mod_clamav_config_rec *rec,
	ap_filter_t *f) {
    mod_clamav_ctx		*ctx;
    int			retries;
#ifdef HAVE_MKSTEMP
    mode_t		mask;
#else
    char		*tmpfilename;
#endif /* HAVE_MKSTEMP */

    /* get the module record						*/
    if (NULL == rec) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] no dir config for clamav", (int)getpid());
	return APR_ENODIR;
    }

    /* initialize the clamav filter context				*/
    if (NULL == f->ctx) {
	/* increment the request counter				*/
	if (rec->shm) {
	    mod_clamav_lock(rec, f->r);
	    rec->stats->requests++;
	    mod_clamav_unlock(rec, f->r);
	}

	/* allocate a new context structure 				*/
	f->ctx = apr_pcalloc(f->r->pool, sizeof(mod_clamav_ctx));
	ctx = (mod_clamav_ctx *)f->ctx;
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] ctx alloced", (int)getpid());

	/* find out whether it is safe to bypass this request, if so,	*/
	/* we don't need the rest of the structure and can safely skip	*/
	/* its initialization						*/
	if (0 != (ctx->bypass = mod_clamav_safe_to_bypass(rec, f)))
	    return APR_SUCCESS;
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] not bypassed", (int)getpid());

	/* create additional data structures				*/
	ctx->bb = apr_brigade_create(f->r->pool, f->c->bucket_alloc);
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] new output brigade created", (int)getpid());

	ctx->module = rec;
	ctx->last_trickle = time(NULL);
	ctx->trickle_offset = 0;
	ctx->bytes = 0;

	/* retry up to five times to create a temporary file		*/
	for (retries = 0; retries < 5; retries++) {
#ifdef HAVE_MKSTEMP
	    /* allocate a temporary filename template			*/
	    ctx->filename = apr_pstrcat(f->r->pool,
		(rec->tmpdir) ? rec->tmpdir : "/tmp", "/clamXXXXXX", NULL);

	    /* create the file and assert access restrictions		*/
	    mask = umask(077);
	    ctx->file = mkstemp(ctx->filename);
	    umask(mask);
	    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] temporary filename is %s", (int)getpid(), ctx->filename);
#else /* HAVE_MKSTEMP */
	    /* allocate a temporary filename				*/
	    tmpfilename = tempnam((rec->tmpdir) ? rec->tmpdir : "/tmp", "clam");
	    ctx->filename = apr_pstrdup(f->r->pool, tmpfilename);
	    free(tmpfilename);
	    ctx->file = open(ctx->filename, O_CREAT|O_TRUNC|O_RDWR, 0600);
	    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] temporary filename is %s", (int)getpid(), ctx->filename);
#endif /* HAVE_MKSTEMP */
	    if (ctx->file >= 0) {
		/* during the download, we use strict permissions,	*/
		/* which we will open up just for the virus scan.	*/
		fchmod(ctx->file, S_IRUSR|S_IWUSR);
		break;	/* stop retrying temporary file creation */
	    }
	}

	/* we open the file read/write, so we can read from the file	*/
	/* without a need to reopen it					*/
	if (ctx->file < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot open temporary file: %s", (int)getpid(),
		strerror(errno));
		return APR_EGENERAL;
	}
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] method %s request for %s", (int)getpid(),
	    (NULL == f->r->method) ? "(null)" : f->r->method,
	    (NULL == f->r->uri) ? "(null)" : f->r->uri);
    }

    /* check for an initialized clamav scan engine, if it is 		*/
    /* initialized, we can skip its initialization			*/
    if (mod_clamav_engine_ready(rec, f))
	return APR_SUCCESS;
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] clamav engine initialization required", (int)getpid());

    /* initialize Clamav engine						*/
    if (mod_clamav_engine_init(rec, f)) {
	return APR_EGENERAL;
    }

    /* if we get to this point, we can be sure that the clamav engine	*/
    /* has been configured correctly					*/
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] initialization of clamav engine complete", (int)getpid());
    return APR_SUCCESS;
} /* mod_clamav_initialize */

/* perform a virus scan */
static int	mod_clamav_scanfile(mod_clamav_config_rec *rec, ap_filter_t *f,
	const char **virname, apr_size_t *len) {

    int			s, l, bytes;
    char		*repl, *code, *v;
    mod_clamav_ctx	*ctx = (mod_clamav_ctx *)f->ctx;
    FILE		*fp;
    char		scancmd[2048];
    int			badreply = (rec->acceptdaemonproblem)
				? CL_CLEAN : CL_VIRUS;
			/* using this return code has the effect of 
			   accepting all files with which the daemon had
			   a problem */

    switch (rec->mode) {
    case MOD_CLAMAV_LOCAL:
	/* virus scan using local clamav library */
#ifdef CL_SCAN_ARCHIVE
	return cl_scanfile(ctx->filename, virname, (unsigned long *)len,
	    rec->common->local->engine, CL_SCAN_ARCHIVE);
#else
	{
	    struct cl_scan_options scanoptions = {0, CL_SCAN_PARSE_ARCHIVE, 0, 0, 0};
	    return cl_scanfile(ctx->filename, virname, (unsigned long *)len,
			       rec->common->local->engine, &scanoptions);
	}
#endif
	break;
    case MOD_CLAMAV_DAEMON:
	/* set reasonable defaults for virname */
	*virname = "daemon connection problem"; *len = 25;

	/* connect to daemon */
	s = socket(rec->common->daemon->domain, SOCK_STREAM, 0);
	if (s < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot create socket: %s (%d)", (int)getpid(),
		strerror(errno), errno);
	    return badreply;
	}
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] daemon socket created: %d", (int)getpid(), s);
	if (connect(s, rec->common->daemon->s, rec->common->daemon->len) < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot connect: %s (%d)", (int)getpid(), strerror(errno),
		errno);
	    close(s);
	    return badreply;
	}
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] connected to daemon", (int)getpid());

	/* just before we do the scan, we change permissions */
	fchmod(ctx->file, rec->perms);

	/* send scanning request */
	bytes = snprintf(scancmd, sizeof(scancmd), "SCAN %s\n", ctx->filename);
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] sending scan command %s", (int)getpid(), scancmd);

	if (bytes != write(s, scancmd, bytes)) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] failed to write scan command: %s", (int)getpid(),
		strerror(errno));
	}

	/* it seems to be easier to read the reply using stdio, so we
	   fdopen the socket */
	if (NULL == (fp = fdopen(s, "r"))) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot fdopen: %s (%d)", (int)getpid(), strerror(errno),
		errno);
	    close(s);
	    fchmod(ctx->file, S_IRUSR|S_IWUSR);
	    return badreply;
	}

	/* read until we meet end of file */
	l = strlen(ctx->filename) + 1024;
	repl = (char *)apr_palloc(f->r->pool, l);
	memset(repl, 0, l);
	if (repl != fgets(repl, l, fp)) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] cannot read reply: %s (%d)", (int)getpid(),
		strerror(errno), errno);
	    fclose(fp);
	    fchmod(ctx->file, S_IRUSR|S_IWUSR);
	    return badreply;
	}
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] got scan reply: %s", (int)getpid(), (repl) ? repl : "(null)");
	fchmod(ctx->file, S_IRUSR|S_IWUSR);
	fclose(fp);

	/* elementary checks on the reply */
	if (repl == NULL) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] NULL scan reply", (int)getpid());
	    return badreply;
	}

	/* check that the reply string is not empty. Since we'll be using
	   a pointer to the last character later, we compute it now */
	l = strlen(repl) - 1;
	if (l < 0) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
		"[%d] short scan reply %d", (int)getpid(), l);
	    return badreply;
	}

	/* as long as we are inside the string, and have \n and \r characters
	   at the end, we keep removing them */
	while ((l >= 0) && ((repl[l] == '\n') || (repl[l] == '\r'))) {
	    repl[l] = '\0';	/* remove the last character */
	    l--;		/* go to the last but one character */
	}

	/* last word: FOUND or ERROR */
	code = strrchr(repl, ' ');
	if (code == repl) {
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r, "[%d] reply has "
		"only one word: %s\n", (int)getpid(), repl);
	    return badreply;
	} 
	*code = '\0'; code++;
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] reply code: %s", (int)getpid(), code);
	if (0 == strcmp(code, "OK"))
	    return CL_CLEAN;
	if (0 == strcmp(code, "ERROR"))
	    return CL_VIRUS;
	if (0 == strcmp(code, "FOUND")) {
	    /* last but one word: virus name */
	    v = strrchr(repl, ' ');
	    if (v == NULL) {
		*virname = "unknown virus";
	    } else {
		*virname = (char *)apr_pstrdup(f->r->pool, v + 1);
	    }
	    *len = strlen(*virname);
	    return CL_VIRUS;
	}

	/* if nothing bad has happend, then the file must be clean 	*/
	return CL_CLEAN;
	break;
    }
    /* this should not happen, but if it does, accept the file 		*/
    return CL_CLEAN;
} /* mod_clamav_scan_file */

/*
 * mod_clamav_cleanup
 *
 * remove temperary files etc
 */
static void	mod_clamav_cleanup(mod_clamav_config_rec *rec, ap_filter_t *f) {
    mod_clamav_ctx	*ctx = (mod_clamav_ctx *)f->ctx;

    /* when done, remove the temporary file 				*/
    if (NULL != ctx->filename) {
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] unlinking %s", (int)getpid(), ctx->filename);
	close(ctx->file);
	unlink(ctx->filename); ctx->filename = NULL;
    }
} /* mod_clamav_cleanup */

/*
 * mod_clamav_send_trickle
 *
 * send a byte every now and then, so that the client does not turn
 * impatient
 */
static apr_status_t	mod_clamav_sendtrickle(mod_clamav_config_rec *rec,
	ap_filter_t *f) {
    time_t		now;
    mod_clamav_ctx	*ctx;
    int			file, bytes;
    char		*str;
    apr_status_t	rc;

    ctx = (mod_clamav_ctx *)f->ctx;
    time(&now);
    if ((now - ctx->last_trickle) < rec->trickle_interval)
	return APR_SUCCESS;
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] trickle time for request %s has arrived", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri);

    /* even if we fail to send the trickle, we just remember the	*/
    /* failure								*/
    ctx->last_trickle = now;

    /* allocate a buffer on the stack which is certailnly large enough	*/
    str = (char *)alloca(rec->trickle_size);

    /* read a suitable number of bytes from the file			*/
    if ((file = open(ctx->filename, O_RDONLY)) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot open file %s: %s (%d)", (int)getpid(), ctx->filename,
	    strerror(errno), errno);
	return APR_SUCCESS;
    }
    if (lseek(file, ctx->trickle_offset, SEEK_SET) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot seek file %s: %s (%d)", (int)getpid(), ctx->filename,
	    strerror(errno), errno);
	close(file);
	return APR_SUCCESS;
    }
    if (0 >= (bytes = read(file, str, rec->trickle_size))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot read file %s: %s (%d)", (int)getpid(), ctx->filename,
	    strerror(errno), errno);
	close(file);
	return APR_SUCCESS;
    }

    /* send it to the client as one bucket followed by a flush bucket	*/
    rc = ap_fwrite(f->next, ctx->bb, str, bytes);
    ctx->trickle_offset += bytes;
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] sending %d trickle bytes starting at %d", (int)getpid(), bytes,
	(int)ctx->trickle_offset);
    APR_BRIGADE_INSERT_TAIL(ctx->bb,
	apr_bucket_flush_create(f->c->bucket_alloc));
    if (rc != APR_SUCCESS) {
	ap_pass_brigade(f->next, ctx->bb);
    } else {
	rc = ap_pass_brigade(f->next, ctx->bb);
    }
    if (f->r->connection->aborted)
	rc = APR_ECONNABORTED;
    if (rc != APR_SUCCESS)
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] sending trickle failed: %d", (int)getpid(), rc);

    /* close the file, file descriptor leak found by Jerome Auge */
    close(file);

    return rc;
} /* mod_clamav_send_trickle */

/*
 * clamav_flush
 *
 * auxiliary function to close the conection to the client
 */
static apr_status_t	mod_clamav_flush(mod_clamav_config_rec *rec,
	ap_filter_t *f) {
    mod_clamav_ctx	*ctx = (mod_clamav_ctx *)f->ctx;

    /* send a flush bucket tot he client 				*/
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] sending eos bucket for request %s", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri);
    APR_BRIGADE_INSERT_TAIL(ctx->bb,
	apr_bucket_flush_create(f->c->bucket_alloc));
    return ap_pass_brigade(f->next, ctx->bb);
} /* mod_clamav_flush */

/*
 * mod_clamav_endfile
 *
 * auxiliary function to close the conection to the client
 */
static apr_status_t	mod_clamav_endfile(mod_clamav_config_rec *rec,
	ap_filter_t *f) {
    mod_clamav_ctx	*ctx = (mod_clamav_ctx *)f->ctx;
    apr_status_t	rc;

    /* send a flush bucket tot he client 				*/
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] sending eos bucket for request %s", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri);
    APR_BRIGADE_INSERT_TAIL(ctx->bb,
	apr_bucket_eos_create(f->c->bucket_alloc));
    rc = ap_pass_brigade(f->next, ctx->bb);

    /* after this we will certainly no longer need the temporary file	*/
    mod_clamav_cleanup(rec, f);

    return rc;
} /* mod_clamav_endfile */

/*
 * mod_clamav_record_aborted
 *
 * auxiliary function to handle an abortet connection
 */
static void	mod_clamav_record_aborted(mod_clamav_config_rec *rec,
	ap_filter_t *f) {

    mod_clamav_set_status_note(rec, f, MOD_CLAMAV_ABORTED_STATUS,
	"client has disconnected", NULL);
    ap_log_rerror(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, 0, f->r,
	"[%d] client requesting %s has disconnected", (int)getpid(),
	(NULL == f->r->uri) ? "(null)" : f->r->uri);
    if (rec->shm) {
	mod_clamav_lock(rec, f->r);
	rec->stats->aborted++;
	mod_clamav_unlock(rec, f->r);
    }
    mod_clamav_cleanup(rec, f);
} /* mod_clamav_record_aborted */

/*
 * mod_clamav_check_file
 *
 * check the file we already have for viruses
 */
static apr_status_t	mod_clamav_check_file(mod_clamav_config_rec *rec,
	ap_filter_t *f) {
    const char		*virname;
    struct rusage	begin, end;
    int			ret;
    apr_size_t		len;
    mod_clamav_ctx	*ctx = (mod_clamav_ctx *)f->ctx;
    struct stat		sb;
    off_t		size = 0;
    
    /* perform virus scan						*/
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] initiating virus check in file %s", (int)getpid(),
	ctx->filename);
    virname = NULL;
    getrusage(RUSAGE_SELF, &begin);
    ret = mod_clamav_scanfile(rec, f, &virname, &len);
    getrusage(RUSAGE_SELF, &end);

    /* log the virus checking achievement				*/
    if (stat(ctx->filename, &sb) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot stat file %s: %s (%d)", (int)getpid(),
	    ctx->filename, strerror(errno), errno);
    } else {
	size = sb.st_size;
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] virus checked %s (%s), %.0f bytes, in %.0fms",
	    (int)getpid(), (NULL == f->r->uri) ? "(null)" : f->r->uri,
	    f->r->content_type, (double)size,
	    1000. * mod_clamav_cpu(rec, &begin, &end));
    }

    /* perform the statistics update					*/
    if (rec->shm) {
	mod_clamav_lock(rec, f->r);
	rec->stats->cpu += mod_clamav_cpu(rec, &begin, &end);
	rec->stats->checked++;
	rec->stats->totalsize += size;
	if (rec->stats->maxsize < size)
	    rec->stats->maxsize = size;
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] statistics: %d+%d bytes, %d checked", (int)getpid(),
	    (int)rec->stats->totalsize, (int)size,
	    (int)rec->stats->checked);
	mod_clamav_unlock(rec, f->r);
     }

    /* if we have found a virus, we also have to register it in the 	*/
    /* statistics 							*/
    if (CL_VIRUS == ret) {
	int	is_ip;
	/* remember this virus */
	if (rec->shm) {
	    mod_clamav_lock(rec, f->r);
	    rec->stats->viruses++;
	    if (rec->stats->maxvirussize < size)
		rec->stats->maxvirussize = size;
	    mod_clamav_virus_record(rec->stats,
		(NULL == f->r->uri) ? "(null)" : f->r->uri,
		ap_get_remote_host(f->r->connection,
		    f->r->per_dir_config, REMOTE_NAME, &is_ip),
		    virname, size);
	    mod_clamav_unlock(rec, f->r);
	}

	/* log the virus we have found */
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] virus %s found in request %s", (int)getpid(),
	    virname,
	    (NULL == f->r->uri) ? "(null)" : f->r->uri);

	/* if we have not sent any data to the client yet, we can
	   generate a HTML page informing about the virus */
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] trickle bytes sent so far: %d (< 2GB)", (int)getpid(),
		(int)ctx->trickle_offset);
	if (ctx->trickle_offset == 0) {
	    mod_clamav_set_status_note(rec, f, MOD_CLAMAV_INFECTED_STATUS,
		"client notified", virname);
	    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"sending virus info to client");
	    mod_clamav_virus_info(rec, f, virname);
	} else {
	    mod_clamav_set_status_note(rec, f, MOD_CLAMAV_INFECTED_STATUS,
		apr_psprintf(f->r->pool, "client NOT notified, "
		    "%d bytes (<2G) sent so far", (int)ctx->trickle_offset),
		    virname);
	}
    }

    /* return the result of the virus scanner, so we know what has	*/
    /* happend from the caller						*/
    return ret;
} /* mod_clamav_check_file */

/*
 * mod_clamav_send_file
 *
 * send all the data we have in the temporary file to the client, and
 * clean the temporary file up
 */
static apr_size_t	mod_clamav_send_file(mod_clamav_config_rec *rec,
	ap_filter_t *f) {
    mod_clamav_ctx	*ctx;
    int			bytes;
    apr_size_t		len = 0;
    size_t		totalsize = 0;
    apr_status_t	rc = APR_SUCCESS;
    char		buffer[2048];
    struct stat		sb;

    ctx = (mod_clamav_ctx *)f->ctx;
    /* send the remaining data as bunch of buckets to the, this is done	*/
    /* as follows							*/
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] sending file contents to client", (int)getpid());

    /* find out how many bytes there currently are in the file		*/
    if (fstat(ctx->file, &sb) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot stat file %s", (int)getpid(), ctx->filename);
    } else {
	totalsize = sb.st_size;
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] expect %ld bytes", (int)getpid(), totalsize);
    }

    /* seek to the current trickle position 				*/
    if (lseek(ctx->file, ctx->trickle_offset, SEEK_SET) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot seek to %d: %s (%d)\n", (int)getpid(),
	    (int)ctx->trickle_offset, strerror(errno), errno);
    }

    /* while we can write to the client, and while there is data, read	*/
    /* a buffer full of data an write it to the client			*/
    while ((rc == APR_SUCCESS) &&
	(bytes = read(ctx->file, buffer, sizeof(buffer))) > 0) {
	/* log if there is a difference between what we asked for and	*/
	/* what we got							*/
	if (bytes != sizeof(buffer))
		mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		    "[%d] read %d bytes from %s", (int)getpid(), bytes,
	    ctx->filename);
	len += bytes;
	rc = ap_fwrite(f->next, ctx->bb, buffer, bytes);
    }
    if (bytes < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] read from descriptor %d failed: %s (%d)", (int)getpid(),
	    ctx->file, strerror(errno), errno);
    }
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] file %s written to client (%ld bytes)", (int)getpid(),
	ctx->filename, len);

    /* send a flush down the filter chaing				*/
    mod_clamav_flush(rec, f);

    /* check whether we have seen all the file				*/
    if (totalsize > 0) {
	if ((totalsize - ctx->trickle_offset) != len) {
	    mod_clamav_set_status_note(rec, f, MOD_CLAMAV_FAILED_STATUS,
		apr_psprintf(f->r->pool, "sent %d of %u bytes",
		    (int)(len + ctx->trickle_offset), (unsigned int)totalsize),
		    NULL);
	    ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, f->r,
		"[%d] not all the file sent to the client: %d != %d\n",
		(int)getpid(), (int)(totalsize - ctx->trickle_offset),
		(int)len);
	} else {
	    mod_clamav_set_status_note(rec, f, MOD_CLAMAV_PASSED_STATUS,
		NULL, NULL);
	    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] sent size matches totalsize %u", (int)getpid(),
		(unsigned int)len);
	}
    } else {
	mod_clamav_set_status_note(rec, f, MOD_CLAMAV_PASSED_STATUS, NULL, NULL);
    }

    /* set the trickle_offset to the current offset, as if we had sent	*/
    /* all the stuff as trickles					*/
    ctx->trickle_offset = totalsize;

    /* clean up the temporary file					*/
    mod_clamav_cleanup(rec, f);

    /* return the number of bytes sent to the client 			*/
    return len;
} /* mod_clamav_send_file */

/*
 * mod_clamav_save_block
 *
 * utility function to write a block to the temporary file
 */
static int	mod_clamav_save_block(mod_clamav_config_rec *rec,
	ap_filter_t *f, const char *str, int len) {
    mod_clamav_ctx	*ctx;
    int			bytes;

    ctx = (mod_clamav_ctx *)f->ctx;
    /* make sure to position the file at the end */
    if (lseek(ctx->file, 0, SEEK_END) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r, "[%d] cannot seek to eof",
	    (int)getpid());
	return 0;
    }

    /* now write the block to the file					*/
    bytes = write(ctx->file, str, len);
    if (bytes != len) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] error during write to %s", (int)getpid(),
	    ctx->filename);
    }
    ctx->bytes += bytes;
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] %d bytes written to tmp file %s", (int)getpid(), bytes,
	ctx->filename);
    return bytes;
} /* mod_clamav_save_block */

/*
 * mod_clamav_filter
 *
 * the apache2 filter function for clamav based virus scanning
 */
apr_status_t	mod_clamav_filter(ap_filter_t *f, apr_bucket_brigade *bb) {
    mod_clamav_ctx		*ctx;
    mod_clamav_config_rec	*rec;
    const char			*str;
    int				bytes;
    apr_size_t			len;
    apr_bucket			*e;
    apr_status_t		rc = APR_SUCCESS;

    /* get the configu record */
    rec = mod_clamav_get_module_config(f->r);
    
    /* initialize (if necessary) the virus filter			*/
    if (APR_SUCCESS != (rc = mod_clamav_initialize(rec, f))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "clamav initialization failed");
	mod_clamav_set_status_note(rec, f, MOD_CLAMAV_FAILED_STATUS,
	    "could not initialize clamav", NULL);
	return rc;
    }
    ctx = (mod_clamav_ctx *)f->ctx;

    /* if it is safe to bypass the virus checker, just pass the bucket	*/
    /* brigade to the next filter					*/
    if (ctx->bypass) {
	return ap_pass_brigade(f->next, bb);
    }

    /* every data bucket that comes in must be written to the file	*/
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] scanning bucket brigade", (int)getpid());
    for (e = APR_BRIGADE_FIRST(bb); e != APR_BRIGADE_SENTINEL(bb);
	e = APR_BUCKET_NEXT(e)) {
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] bucket type: %s (uri %s)", (int)getpid(), e->type->name,
	    (NULL == f->r->uri) ? "(null)" : f->r->uri);
	/* read the data in the bucket					*/
	apr_bucket_read(e, &str, &len, APR_NONBLOCK_READ);
	if (len > 0) {
	    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] perform pattern checks", (int)getpid());
	    if (ctx->bytes == 0) {
		/* perform checking of the patterns			*/
		mod_clamav_safepattern *p =
		    (mod_clamav_safepattern *)(rec->safepatterns->elts);
		int total = rec->safepatterns->nelts;
		int i,j;

		for (i = 0; i < total; i++) {
		    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
			"[%d] check pattern %s", (int)getpid(), p[i].tag);
		    /* check for mode -1, i.e. ??? */
		    if (p[i].mode == -1) {
			/* the check only makes sense if we have enough data */
			if(len >= sizeof(p->pattern)) {
			    for (j = 0; j < sizeof(p->pattern); j++) {
				/* is this text? */
				if(str[j] >= 0x20 && str[j] <= 0x7e);
				else if(str[j]=='\t');
				else if(str[j]=='\r');
				else if(str[j]=='\n');
				else break;
			    }
			    /* if we have not left up to the end of the	*/
			    /* buffer, it must be text			*/
			    if (j == sizeof(p->pattern)) {
				ctx->bypass = 1;
				mod_clamav_cleanup(rec, f);
				mod_clamav_set_bypassed_note(rec, f,
				    apr_psprintf(f->r->pool,
					"'%s' allows bypass", p[i].tag));
				mod_clamav_ap_log_rerror(APLOG_MARK,
				    APLOG_DEBUG, 0, f->r,
				    "[%d] %s allows bypass for %s",
				    (int)getpid(), p[i].tag,
				    (f->r->uri) ? f->r->uri : "(null)");
				return ap_pass_brigade(f->next, bb);
			    }
			}
		    } else {
			/* check for binary stuff, need at least as 	*/
			/* many bytes as the mode indicates		*/
			if(len >= p[i].mode) {
			    /* compare the data to the pattern */
			    for (j = 0; j < p[i].mode; j++) {
				if ((str[j] & p[i].mask[j]) != p[i].pattern[j])
				    break;
			    }
			    /* if we have not left prematurely, then we	*/
			    /* have a match				*/
			    if (j == p[i].mode) {
				ctx->bypass = 1;
				mod_clamav_cleanup(rec, f);
				mod_clamav_set_bypassed_note(rec, f,
				    apr_psprintf(f->r->pool,
					"'%s' allows bypass", p[i].tag));
				mod_clamav_ap_log_rerror(APLOG_MARK,
				    APLOG_DEBUG, 0, f->r,
				    "[%d] %s allows bypass for %s",
				    (int)getpid(), p[i].tag,
				    (f->r->uri) ? f->r->uri : "(null)");
				return ap_pass_brigade(f->next, bb);
			    }
			}
		    }
		}
	    }

	    /* there really is something to write			*/
	    if (len != (bytes = mod_clamav_save_block(rec, f, str, len))) {
		    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
			"[%d] %d bytes written to tmp file %s",
			(int)getpid(), bytes, ctx->filename);
	    }
	}

	/* should give EOS and FLUSH buckets special treatment. But as	*/
	/* noted below, Flush buckets don't really help, so we just drop*/
	/* them, not without a debug log message though			*/
	if ((!APR_BUCKET_IS_EOS(e)) && (!APR_BUCKET_IS_FLUSH(e))) {
	    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] data bucket found", (int)getpid());
	    apr_bucket_delete(e); /* removes any non-EOS or non-FLUSH buckets */
	} else {
	    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
		"[%d] bucket of type %s found", (int)getpid(),
		APR_BUCKET_IS_EOS(e) ? "End of Stream" : "Flush");
	    if (APR_BUCKET_IS_FLUSH(e)) {
		apr_bucket_delete(e); /* removes  any FLUSH buckets */
	    }
	}
	/* at this point, all non-EOS buckets have been removed		*/
    }

    /* now check whether we should send a trickle byte, but sending	*/
    /* a trickle may fail caused by the client having aborted the 	*/
    /* the connection, so we return an error				*/
    if (APR_SUCCESS != mod_clamav_sendtrickle(rec, f)) {
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
	    "[%d] trickle detects connection abort", (int)getpid());
	mod_clamav_record_aborted(rec, f);
	return APR_ECONNABORTED;
    }

    /* if we have read up to the size limit already, then we check	*/
    /* the piece we have and send it to the client, and switch to	*/
    /* bypass mode							*/
    if ((rec->sizelimit > 0) && (rec->sizelimit <= ctx->bytes)){
	/* perform the check of the part we already have		*/
	if (CL_VIRUS == mod_clamav_check_file(rec, f)) {
	    mod_clamav_endfile(rec, f);
	    return APR_SUCCESS;
	}

	/* send the file to the client					*/
	mod_clamav_send_file(rec, f);

	/* switch to bypass mode					*/
	mod_clamav_set_bypassed_note(rec, f, "size limit reached");
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
	    "bypassing check of %s due to size limit %d",
	    f->r->uri, rec->sizelimit);
	ctx->bypass = 1;

	/* but we should hand over the bucket brigade to the next	*/
	/* as if we had been bypassing already at the beginning		*/
	return ap_pass_brigade(f->next, bb);
    }

    /* at this point we only have EOS buckets left			*/
    for (e = APR_BRIGADE_FIRST(bb); e != APR_BRIGADE_SENTINEL(bb);
	e = APR_BUCKET_NEXT(e)) {
	/* if we have met an EOS bucket, we should trigger the virus	*/
	/* scanning							*/
	if (APR_BUCKET_IS_EOS(e)) {
	    /* check for viruses and update the statistics		*/
	    if (CL_VIRUS == mod_clamav_check_file(rec, f)) {
		/* close the connection to the client */
		mod_clamav_endfile(rec, f);
		return APR_SUCCESS;
	    }

	    /* send the remaining data as bunch of buckets to the 	*/
	    /* client							*/
	    mod_clamav_send_file(rec, f);
	    rc = mod_clamav_endfile(rec, f);
	    if (rc != APR_SUCCESS)
		mod_clamav_record_aborted(rec, f);
	    return rc;
	} else {
	    /* for the record: we ignore FLUSH buckets, they don't help	*/
	    /* to detect that the connection was aborted		*/
	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
		"[%d] non-EOS bucket found when only EOS buckets should be "
		"left", (int)getpid());
	}
	/* for the record: we ignore FLUSH buckets, they don't help to	*/
	/* detect that the connection was aborted			*/
    }
    mod_clamav_set_status_note(rec, f, MOD_CLAMAV_PASSED_STATUS,
	"reached end of clamav_filter", NULL);
    apr_brigade_cleanup(bb);
    return APR_SUCCESS;
} /* mod_clamav_filter */

