/* 
**  handler.c -- handler related stuff for the mod_clamav module
**
**  (c) 2003-2009 Dr. Andreas Mueller, Beratung und Entwicklung
**
**  This module is distributed under the GNU General Public License (GPL).
**  See the file COPYING for details.
**
**  $Id: handler.c,v 1.1 2009/04/10 23:32:37 afm Exp $
*/ 
#include <mod_clamav.h>
#include <mod_clamav_version.h>

void	mod_clamav_virus_record(mod_clamav_stats *stats,
	const char *uri, const char *req, const char *virus, off_t size) {
    /* find the index of the entry we are about to create */
    if (++stats->last >= VIRUS_LIST_LENGTH) stats->last = 0;
    stats->lastviruses[stats->last].pid = getpid();
    stats->lastviruses[stats->last].when = apr_time_now();
    stats->lastviruses[stats->last].size = size;

    /* reset the string fields */
    memset(stats->lastviruses[stats->last].uri, 0, VIRUS_URI_LENGTH);
    memset(stats->lastviruses[stats->last].req, 0, VIRUS_REQ_LENGTH);
    memset(stats->lastviruses[stats->last].virus, 0, VIRUS_NAME_LENGTH);

    /* remember uri and virus name */
    strncpy(stats->lastviruses[stats->last].uri, uri, VIRUS_URI_LENGTH - 1);
    strncpy(stats->lastviruses[stats->last].req, req, VIRUS_REQ_LENGTH - 1);
    strncpy(stats->lastviruses[stats->last].virus, virus, VIRUS_NAME_LENGTH -1);
} /* clamav_virus_record */

static void	mod_clamav_set_note(ap_filter_t *f, const char *note,
	const char *content) {
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] %s - %s", (int)getpid(), note, content);
    apr_table_set(f->r->notes, note, content);
}

void	mod_clamav_set_status_note(mod_clamav_config_rec *rec,
	ap_filter_t *f, const char *status, const char *details,
	const char *virusname) {
    apr_table_t	*notes = f->r->notes;
    const char	*longstatus = NULL;
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] setting status message: status=%s, details=%s, virusname=%s",
	(int)getpid(),
	(status) ? status : "(no status)",
	(details) ? details : "(no details)",
	(virusname) ? virusname : "(no virusname)");

    if (!rec->extendedlogging) {
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] extended logging not enabled", (int)getpid());
	return;
    }

    /* set the individual notes						*/
    mod_clamav_set_note(f, MOD_CLAMAV_STATUS_NOTE,
	(status) ? status : "-");
    mod_clamav_set_note(f, MOD_CLAMAV_DETAILS_NOTE,
	(details) ? details : "-");
    mod_clamav_set_note(f, MOD_CLAMAV_VIRUSNAME_NOTE,
	(virusname) ? virusname : "-");

    /* build a long status message from the individual parts		*/
    if (status) {
	longstatus = status;
    }

    /* add the details to the longstatus				*/
    if (details) {
	if (longstatus) {
	    longstatus = apr_psprintf(f->r->pool, "%s, %s", longstatus,
		details);
	} else {
	    longstatus = details;
	}
    }

    /* add the virus name						*/
    if (virusname) {
	if (longstatus) {
	    longstatus = apr_psprintf(f->r->pool, "%s, found virus: %s",
		longstatus, virusname);
	} else {
	    longstatus = virusname;
	}
    }

    /* add the status to the notes table				*/
    mod_clamav_set_note(f, MOD_CLAMAV_LONGSTATUS_NOTE,
	longstatus ? longstatus : "(no status)");
} /* clamav_set_status_note */

/* create HTML page informing about virus */
int	mod_clamav_virus_info(mod_clamav_config_rec *rec, ap_filter_t *f,
	const char *virname) {
    char	infopage[HUGE_STRING_LEN];
    mod_clamav_ctx	*ctx = (mod_clamav_ctx *)f->ctx;
    /* set content type header to text/html */
    f->r->content_type = "text/html";      

    /* format page output */
    if (!f->r->header_only) {
	apr_table_setn(f->r->headers_out, "Cache-Control", "no-cache");
	apr_table_setn(f->r->headers_out, "Pragma", "no-cache");
	if (rec->message) {
	    char	*msg = rec->message;
	    const char	*p;
	    int		i = 0, j = 0;
	    while (i < sizeof(infopage) - 1 && msg[j]) {
		if (msg[j] != '%') {
		    infopage[i++] = msg[j++];
		} else {
		    j++;
		    switch (msg[j]) {
		    case 'v':
			j++;
			p = virname;
			while (i < sizeof(infopage) - 1 && *p) {
			    infopage[i++] = *p++;
			}
			break;
		    case 'i':
			j++;
			p = "mod_clamav ";
			while (i < sizeof(infopage) - 1 && *p) {
			    infopage[i++] = *p++;
			}
			p = mod_clamav_version;
			while (i < sizeof(infopage) - 1 && *p) {
			    infopage[i++] = *p++;
			}
			break;
		    case 'u':
			j++;
			p = (NULL == f->r->uri) ? "(null)" : f->r->uri;
			while (i < sizeof(infopage) - 1 && *p) {
			    infopage[i++] = *p++;
			}
			break;
		    case '\0':
			break;
		    default:
			infopage[i++] = msg[j++];
		    }
		}
	    }
	    infopage[i] = '\0';
	} else {
	    snprintf(infopage, sizeof(infopage),
		"<html>\n"
		"<head>\n"
		"<title>Virus found</title>\n"
		"</head>\n"
		"<body>\n"
		"<h1>Virus found</h1>\n"
		"<p><tt>mod_clamav</tt> version %s found the virus</p>"
		"<p><pre><b>%s</b></pre></p>"
		"<p>while downloading <i>%s</i>."
		"The transfer has been aborted.</p>\n"
		"<hr />&copy; mod_clamav %s\n"
		"</body>\n"
		"</html>\n",
	    mod_clamav_version, virname,
	    (NULL == f->r->uri) ? "(null)" : f->r->uri, mod_clamav_version);
	}
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] sending %ld bytes to client", (int)getpid(), strlen(infopage));
	ap_fwrite(f->next, ctx->bb, infopage, strlen(infopage));
    }
    /* this never happens, but we want the compiler to be happy		*/
    return 0;
} /* clamav_virus_info */

/* The content handler */
void	mod_clamav_display_virus(mod_clamav_config_rec *rec, request_rec *r,
	mod_clamav_virusinfo *v) {
    char		when[32];
    apr_size_t		ret;
    apr_time_exp_t	t;
    /* if pid is zero, the record is not initialized */
    if (v->pid == 0) return;

    ap_rputs("  <tr>\n", r);
    ap_rprintf(r, "    <td>%d</td>\n", (int)(v->pid));
    apr_time_exp_lt(&t, v->when);
    apr_strftime(when, &ret, sizeof(when), "%d %b %Y %H:%M:%S", &t);
    ap_rprintf(r, "    <td>%s</td>\n", when);
    ap_rprintf(r, "    <td>%s</td>\n", v->req);
    ap_rprintf(r, "    <td>%s</td>\n", v->virus);
    ap_rprintf(r, "    <td>%s</td>\n", v->uri);
    ap_rprintf(r, "    <td align=\"right\">%.0f</td>\n", (double)(v->size));
    ap_rputs("  </tr>\n", r);
} /* clamav_display_virus */

/*
 * clamav_handler
 *
 * the clamav statistics content handler
 */
int mod_clamav_handler(request_rec *r) {
    mod_clamav_config_rec	*rec;
    int			i, from;

    /* get the module record						*/
    rec = mod_clamav_get_module_config(r);

    /* make sure the shared memory is initialized			*/
    mod_clamav_shm(rec, r);

    /* initialize the clamav filter context				*/
    if (strcmp(r->handler, "clamav")) {
        return DECLINED;
    }
    r->content_type = "text/html";      

    if (!r->header_only) {
	/* lock the shared memory */
	mod_clamav_lock(rec, r);
	ap_rputs("<html>", r);
	ap_rputs("<head><title>Clamav Module Statistics</title></head>\n", r);
	ap_rputs("<body bgcolor=\"#ffffff\">\n", r);
	ap_rputs("<h1>Clamav Module Status</h1>\n", r);
	ap_rputs("<h2>Configuration</h2>\n", r);
	ap_rputs("<table>\n", r);
	ap_rputs("  <tr>\n", r);
	ap_rprintf(r, "    <td>Version:</td><td>%s</td>\n", mod_clamav_version);
	ap_rputs("  </tr>\n", r);
	ap_rputs("  <tr>\n", r);
	if (rec->mode == MOD_CLAMAV_LOCAL) {
	    ap_rputs("    <td>Scanner:</td><td>local</td>\n", r);
	} else {
	    if (rec->socket)
		ap_rprintf(r, "    <td>Daemon Unix Socket:</td><td>%s</td>\n",
		    rec->socket);
	    else
		ap_rprintf(r, "    <td>Daemon TCP Socket:</td>"
		    "<td>127.0.0.1:%d</td>\n", rec->port);
	}
	ap_rputs("  </tr>\n", r);
	ap_rputs("</table>\n", r);
	ap_rputs("<h2>Statistics</h2>\n", r);
	if (rec->shm) {
	    ap_rputs("<table>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Total requests:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		rec->stats->requests);
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Checked for viruses:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		rec->stats->checked);
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Total data checked:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%.0f</td><td>kB</td>\n",
		(double)(rec->stats->totalsize)/1024.);
	    ap_rputs("  </tr>\n", r);
	    if (rec->stats->checked > 0) {
		ap_rputs("  <tr>\n", r);
		ap_rputs("    <td>Average size of checked requests:</td>\n", r);
		ap_rprintf(r, "    <td align=\"right\">%.f</td><td>bytes</td>\n",
		    (double)rec->stats->totalsize/(double)rec->stats->checked);
		ap_rputs("  </tr>\n", r);
	    }
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Maximum request size:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%.0f</td><td>bytes</td>\n",
		(double)(rec->stats->maxsize));
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Maximum virus size:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%.0f</td><td>bytes</td>\n",
		(double)(rec->stats->maxvirussize));
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Viruses found:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		rec->stats->viruses);
	    ap_rputs("  </tr>\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>Aborted:</td>\n", r);
	    ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		rec->stats->aborted);
	    ap_rputs("  </tr>\n", r);
	    if (rec->mode == MOD_CLAMAV_LOCAL) {
		ap_rputs("  <tr>\n", r);
		ap_rputs("    <td>Total CPU time:</td>\n", r);
		ap_rprintf(r, "    <td align=\"right\">%.3f</td><td>sec</td>\n",
		    rec->stats->cpu);
		ap_rputs("  </tr>\n", r);
		ap_rputs("  <tr>\n", r);
		ap_rputs("    <td>Database reloads:</td>\n", r);
		ap_rprintf(r, "    <td align=\"right\">%ld</td>\n",
		    rec->stats->reloads);
		ap_rputs("  </tr>\n", r);
	    }
	    ap_rputs("</table>\n", r);

	    ap_rprintf(r, "<h2>Last %d viruses found:</h2>\n",
		VIRUS_LIST_LENGTH);

	    ap_rputs("<table border=\"1\">\n", r);
	    ap_rputs("  <tr>\n", r);
	    ap_rputs("    <td>PID</td>\n", r);
	    ap_rputs("    <td>When</td>\n", r);
	    ap_rputs("    <td>Requestor</td>\n", r);
	    ap_rputs("    <td>Virus</td>\n", r);
	    ap_rputs("    <td>Request URI</td>\n", r);
	    ap_rputs("    <td align=\"right\">Size</td>\n", r);
	    ap_rputs("  </tr>\n", r);

	    /* tail portion of the list */
	    from = rec->stats->last + 1;
	    if (from < 1) from = 1;
	    for (i = from; i < VIRUS_LIST_LENGTH; i++)
		mod_clamav_display_virus(rec, r, &rec->stats->lastviruses[i]);

	    /* front portion of the list */
	    for (i = 0; i <= rec->stats->last; i++)
		mod_clamav_display_virus(rec, r, &rec->stats->lastviruses[i]);

	    ap_rputs("</table>\n", r);
	} else {
	    ap_rputs("<p>No statistics available</p>\n", r);
	}

	ap_rputs("<hr />\n", r);
	ap_rputs("&copy; 2003-2009 Dr. Andreas M&uuml;ller, "
	    "Beratung und Entwicklung\n", r);
	ap_rputs("</body>\n", r);
	ap_rputs("</html>\n", r);
	mod_clamav_unlock(rec, r);
    }
    return OK;
} /* clamav_handler */

/*
 * clamav_cpu
 *
 * auxiliary function to compute the cpu time consumed for virus scanning
 */
double	mod_clamav_cpu(mod_clamav_config_rec *rec, struct rusage *begin,
	struct rusage *end) {
    long	sec, usec;
    double	cpu;
    sec = end->ru_utime.tv_sec - begin->ru_utime.tv_sec;
    usec = end->ru_utime.tv_usec - begin->ru_utime.tv_usec;
    cpu = sec + (0.000001 * usec);
    sec = end->ru_stime.tv_sec - begin->ru_stime.tv_sec;
    usec = end->ru_stime.tv_usec - begin->ru_stime.tv_usec;
    return cpu + sec + (0.0000001 * usec);
} /* clamav_cpu */

