/* 
**  init.c -- initialize virus scanner for mod_clamav
**
**  (c) 2003-2009 Dr. Andreas Mueller, Beratung und Entwicklung
**
**  This module is distributed under the GNU General Public License (GPL).
**  See the file COPYING for details.
**
**  $Id: init.c,v 1.1 2009/04/10 23:32:37 afm Exp $
*/ 
#include <mod_clamav.h>

/* initialize the engine (depends on mode) */
static int	mod_clamav_init_local(mod_clamav_config_rec *rec,
	ap_filter_t *f) {
    int	ret;
    unsigned int	signo = 0;
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] local init", (int)getpid());

    /* initializa clam */
    if (0 != (ret = cl_init(CL_INIT_DEFAULT))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot initialize clamav: %s", (int)getpid(),
	    cl_strerror(ret));
	return -1;
    }

    /*  make sure that rec->common->local is allocated			*/
    if (NULL == rec->common->local) {
	rec->common->local = (cl_local_t *)apr_palloc(rec->pool,
	    sizeof(cl_local_t));
	rec->common->local->engine = cl_engine_new();
	rec->common->local->lastreload = 0;
    }

    /* set the size parameters */
    cl_engine_set_num(rec->common->local->engine, CL_ENGINE_MAX_FILES,
	rec->maxfiles);
    cl_engine_set_num(rec->common->local->engine, CL_ENGINE_MAX_FILESIZE,
	rec->maxfilesize);
    cl_engine_set_num(rec->common->local->engine, CL_ENGINE_MAX_RECURSION,
	rec->maxreclevel);

    /* make sure we know in which directory to look for databases	*/
    if (NULL == rec->dbdir) {
	rec->dbdir = cl_retdbdir();
    }

    /* reload the database from the directory				*/
    if (0 != (ret = cl_load(rec->dbdir, rec->common->local->engine,
	&signo, CL_DB_STDOPT))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot load clamav patterns: %s", (int)getpid(),
	    cl_strerror(ret));
	return -1;
    }
    if (rec->shm) {
	mod_clamav_lock(rec, f->r);
	rec->stats->reloads++;
	mod_clamav_unlock(rec, f->r);
    }
    time(&rec->common->local->lastreload); 	/* remember when reloaded */
#if 0
    if (0 != (ret = cl_build(rec->common->local->engine))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r,
	    "[%d] cannot build clamav tree: %s", (int)getpid(),
		cl_strerror(ret));
	cl_free(rec->common->local->engine);
	return -1;
    }
#endif 

    /* if we get to this point, we were successfull initializing the local*/
    /* scanner								*/
    return 0;
} /* mod_clamav_init_local */

static int	mod_clamav_init_daemon(mod_clamav_config_rec *rec, ap_filter_t *f) {
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "[%d] daemon init",
	(int)getpid());
    rec->common->daemon = (cl_daemon_t *)apr_palloc(rec->pool,
	sizeof(cl_daemon_t));
    if (rec->socket) {
	/* initialize daemon connections via a Unix domain socket	*/
	struct sockaddr_un	*sau;
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] init unix domain socket '%s'", (int)getpid(), rec->socket);
	sau = (struct sockaddr_un *)apr_palloc(rec->pool,
	    sizeof(struct sockaddr_un));
	rec->common->daemon->s = (struct sockaddr *)sau;
	rec->common->daemon->len = sizeof(struct sockaddr_un);
	rec->common->daemon->domain = AF_UNIX;
	/* initialize unix domain socket address */
	/* sau->sun_len = strlen(rec->socket); */
	sau->sun_family = PF_UNIX;
	strncpy(sau->sun_path, rec->socket, sizeof(sau->sun_path) - 1);
	sau->sun_path[sizeof(sau->sun_path) - 1] = '\0';
    } else {
	/* initialize daemon connections via a TCP socket		*/
	struct sockaddr_in	*sai;
	sai = (struct sockaddr_in *)apr_palloc(rec->pool,
	    sizeof(struct sockaddr_in));
	rec->common->daemon->s = (struct sockaddr *)sai;
	rec->common->daemon->len = sizeof(struct sockaddr_in);
	rec->common->daemon->domain = AF_INET;
	/* initialize INET domain socket address */
	/* sai->sin_len = sizeof(sai->sin_addr); */
	sai->sin_family = PF_INET;
	sai->sin_port = htons((short)rec->port);
	sai->sin_addr.s_addr = htonl(0x7f000001);
	mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	    "[%d] init inet socket to %s:%d", (int)getpid(),
	    inet_ntoa(sai->sin_addr), rec->port);
    }
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] daemon init complete", (int)getpid());
    return 0;
} /* mod_clamav_init_daemon */

/* initialize the virus scanning engine: either the library inside the
   binary, or using the daemon (prefered) */
int	mod_clamav_engine_init(mod_clamav_config_rec *rec, ap_filter_t *f) {
    /* initialize share memory */
    mod_clamav_shm(rec, f->r);

    /* initialize the engine */
    switch (rec->mode) {
    case MOD_CLAMAV_LOCAL:
	return mod_clamav_init_local(rec, f);
	break;
    case MOD_CLAMAV_DAEMON:
	return mod_clamav_init_daemon(rec, f);
	break;
    default:
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, f->r, "[%d] unknown mode. %d",
	    (int)getpid(), rec->mode);
    }
    mod_clamav_ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
	"[%d] working in mode %d", (int)getpid(), rec->mode);
    return  -1;
} /* mod_clamav_engine_init */

/* check whether the engine is ready. In the case of the local scanner,
   reloads the database if necessary as a side effect */
int	mod_clamav_engine_ready(mod_clamav_config_rec *rec, ap_filter_t *f) {
    time_t	now;
    switch (rec->mode) {
    case MOD_CLAMAV_LOCAL:
	/* the module is not read if the local structure has never been	*/
	/* allocated							*/
	if (NULL == rec->common->local)
	    return 0;
	/* even if the local structure is allocated, it may be the case	*/
	/* that the reload interval has passed, so we may have to reload*/
	/* the database							*/
	if (0 == rec->reload_interval)
	    return 1;	/* always ready, since we are not going to ever	*/
			/* reload the pattern database			*/
	time(&now);
	if ((now - rec->common->local->lastreload) > rec->reload_interval) {
	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, f->r,
		"[%d] database reload necessary after %d seconds",
		(int)getpid(), (int)(now - rec->common->local->lastreload));
	    /* make sure the engine is properly deallocated		*/
	    cl_engine_free(rec->common->local->engine);
	    return 0;
	}
	return 1;
	break;
    case MOD_CLAMAV_DAEMON:
	return (NULL != rec->common->daemon);
	break;
    }
    /* no other state is possible, but how should the compiler know	*/
    /* that? so if we get a different state, will just tell the caller	*/
    /* that we aren't ready, which isn't that wrong after all		*/
    return 0;
} /* mod_clamav_engine_ready */

